#ifndef DISCRETE_FUNCTION_P0_HPP
#define DISCRETE_FUNCTION_P0_HPP

#include <scheme/IDiscreteFunction.hpp>

#include <language/utils/InterpolateItemValue.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshData.hpp>
#include <mesh/MeshDataManager.hpp>
#include <scheme/DiscreteFunctionDescriptorP0.hpp>

template <size_t Dimension, typename DataType>
class DiscreteFunctionP0 : public IDiscreteFunction
{
 public:
  using data_type = DataType;
  using MeshType  = Mesh<Connectivity<Dimension>>;

 private:
  std::shared_ptr<const MeshType> m_mesh;
  CellValue<DataType> m_cell_values;

  DiscreteFunctionDescriptorP0 m_discrete_function_descriptor;

 public:
  ASTNodeDataType
  dataType() const final
  {
    return ast_node_data_type_from<DataType>;
  }

  const CellValue<DataType>&
  cellValues() const
  {
    return m_cell_values;
  }

  std::shared_ptr<const IMesh>
  mesh() const
  {
    return m_mesh;
  }

  const IDiscreteFunctionDescriptor&
  descriptor() const final
  {
    return m_discrete_function_descriptor;
  }

  PUGS_FORCEINLINE
  DataType&
  operator[](const CellId& cell_id) const noexcept(NO_ASSERT)
  {
    return m_cell_values[cell_id];
  }

  friend DiscreteFunctionP0
  operator+(const DiscreteFunctionP0& f, const DiscreteFunctionP0& g)
  {
    Assert(f.mesh() == g.mesh(), "functions are nor defined on the same mesh");
    std::shared_ptr mesh = std::dynamic_pointer_cast<const MeshType>(f.mesh());
    DiscreteFunctionP0 sum(mesh);
    parallel_for(
      mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) { sum[cell_id] = f[cell_id] + g[cell_id]; });
    return sum;
  }

  friend DiscreteFunctionP0
  operator-(const DiscreteFunctionP0& f, const DiscreteFunctionP0& g)
  {
    Assert(f.mesh() == g.mesh(), "functions are nor defined on the same mesh");
    std::shared_ptr mesh = std::dynamic_pointer_cast<const MeshType>(f.mesh());
    DiscreteFunctionP0 difference(mesh);
    parallel_for(
      mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) { difference[cell_id] = f[cell_id] - g[cell_id]; });
    return difference;
  }

  friend DiscreteFunctionP0
  operator*(const DiscreteFunctionP0& f, const DiscreteFunctionP0& g)
  {
    Assert(f.mesh() == g.mesh(), "functions are nor defined on the same mesh");
    std::shared_ptr mesh = std::dynamic_pointer_cast<const MeshType>(f.mesh());
    DiscreteFunctionP0 product(mesh);
    parallel_for(
      mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) { product[cell_id] = f[cell_id] * g[cell_id]; });
    return product;
  }

  template <typename DataType2T>
  friend DiscreteFunctionP0<Dimension, decltype(DataType2T{} * DataType{})>
  operator*(const DiscreteFunctionP0<Dimension, DataType2T>& f, const DiscreteFunctionP0& g)
  {
    Assert(f.mesh() == g.mesh(), "functions are nor defined on the same mesh");
    std::shared_ptr mesh = std::dynamic_pointer_cast<const MeshType>(f.mesh());
    DiscreteFunctionP0<Dimension, decltype(DataType2T{} * DataType{})> product(mesh);
    parallel_for(
      mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) { product[cell_id] = f[cell_id] * g[cell_id]; });
    return product;
  }

  friend DiscreteFunctionP0
  operator*(const double& a, const DiscreteFunctionP0& f)
  {
    std::shared_ptr mesh = std::dynamic_pointer_cast<const MeshType>(f.mesh());
    DiscreteFunctionP0 product(mesh);
    parallel_for(
      mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) { product[cell_id] = a * f[cell_id]; });
    return product;
  }

  friend DiscreteFunctionP0
  operator/(const DiscreteFunctionP0& f, const DiscreteFunctionP0& g)
  {
    Assert(f.mesh() == g.mesh(), "functions are nor defined on the same mesh");
    std::shared_ptr mesh = std::dynamic_pointer_cast<const MeshType>(f.mesh());
    DiscreteFunctionP0 ratio(mesh);
    parallel_for(
      mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) { ratio[cell_id] = f[cell_id] / g[cell_id]; });
    return ratio;
  }

  DiscreteFunctionP0(const std::shared_ptr<const MeshType>& mesh, const FunctionSymbolId& function_id) : m_mesh(mesh)
  {
    using MeshDataType      = MeshData<Dimension>;
    MeshDataType& mesh_data = MeshDataManager::instance().getMeshData(*mesh);

    m_cell_values =
      InterpolateItemValue<DataType(TinyVector<Dimension>)>::template interpolate<ItemType::cell>(function_id,
                                                                                                  mesh_data.xj());
  }

  DiscreteFunctionP0(const std::shared_ptr<const MeshType>& mesh) : m_mesh{mesh}, m_cell_values{mesh->connectivity()}
  {
    ;
  }

  DiscreteFunctionP0(const std::shared_ptr<const MeshType>& mesh, const CellValue<DataType>& cell_value)
    : m_mesh{mesh}, m_cell_values{cell_value}
  {
    ;
  }

  DiscreteFunctionP0(const DiscreteFunctionP0&) noexcept = default;
  DiscreteFunctionP0(DiscreteFunctionP0&&) noexcept      = default;

  ~DiscreteFunctionP0() = default;
};

template <size_t Dimension, size_t ValueDimension>
DiscreteFunctionP0<Dimension, TinyVector<ValueDimension>>
operator*(const TinyMatrix<ValueDimension>& A, const DiscreteFunctionP0<Dimension, TinyVector<ValueDimension>>& f)
{
  using MeshType       = typename DiscreteFunctionP0<Dimension, TinyVector<ValueDimension>>::MeshType;
  std::shared_ptr mesh = std::dynamic_pointer_cast<const MeshType>(f.mesh());
  DiscreteFunctionP0<Dimension, TinyVector<ValueDimension>> product(mesh);
  parallel_for(
    mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) { product[cell_id] = A * f[cell_id]; });
  return product;
}

template <size_t Dimension, size_t ValueDimension>
DiscreteFunctionP0<Dimension, TinyMatrix<ValueDimension>>
operator*(const TinyMatrix<ValueDimension>& A, const DiscreteFunctionP0<Dimension, TinyMatrix<ValueDimension>>& f)
{
  using MeshType       = typename DiscreteFunctionP0<Dimension, TinyMatrix<ValueDimension>>::MeshType;
  std::shared_ptr mesh = std::dynamic_pointer_cast<const MeshType>(f.mesh());
  DiscreteFunctionP0<Dimension, TinyMatrix<ValueDimension>> product(mesh);
  parallel_for(
    mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) { product[cell_id] = A * f[cell_id]; });
  return product;
}

template <size_t Dimension, size_t ValueDimension>
DiscreteFunctionP0<Dimension, TinyMatrix<ValueDimension>>
operator*(const DiscreteFunctionP0<Dimension, TinyMatrix<ValueDimension>>& f, const TinyMatrix<ValueDimension>& A)
{
  using MeshType       = typename DiscreteFunctionP0<Dimension, TinyMatrix<ValueDimension>>::MeshType;
  std::shared_ptr mesh = std::dynamic_pointer_cast<const MeshType>(f.mesh());
  DiscreteFunctionP0<Dimension, TinyMatrix<ValueDimension>> product(mesh);
  parallel_for(
    mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) { product[cell_id] = f[cell_id] * A; });
  return product;
}

template <size_t Dimension, size_t ValueDimension>
DiscreteFunctionP0<Dimension, TinyMatrix<ValueDimension>>
operator*(const DiscreteFunctionP0<Dimension, double>& f, const TinyMatrix<ValueDimension>& A)
{
  using MeshType       = typename DiscreteFunctionP0<Dimension, TinyMatrix<ValueDimension>>::MeshType;
  std::shared_ptr mesh = std::dynamic_pointer_cast<const MeshType>(f.mesh());
  DiscreteFunctionP0<Dimension, TinyMatrix<ValueDimension>> product(mesh);
  parallel_for(
    mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) { product[cell_id] = f[cell_id] * A; });
  return product;
}

template <size_t Dimension, size_t ValueDimension>
DiscreteFunctionP0<Dimension, TinyVector<ValueDimension>>
operator*(const DiscreteFunctionP0<Dimension, double>& f, const TinyVector<ValueDimension>& A)
{
  using MeshType       = typename DiscreteFunctionP0<Dimension, TinyVector<ValueDimension>>::MeshType;
  std::shared_ptr mesh = std::dynamic_pointer_cast<const MeshType>(f.mesh());
  DiscreteFunctionP0<Dimension, TinyVector<ValueDimension>> product(mesh);
  parallel_for(
    mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) { product[cell_id] = f[cell_id] * A; });
  return product;
}

#endif   // DISCRETE_FUNCTION_P0_HPP
